from soundcloud import SoundCloud


def test_get_comments_with_interactions(client: SoundCloud):
    comments = list(client.get_track_comments_with_interactions(1032303631))
    comments = list(filter(lambda comment: comment.comment.id == 1509855118, comments))
    assert comments
    comment = comments[0]
    assert comment.likes >= 1
    assert comment.liked_by_creator
    assert comment.liked_by_user



import itertools

from soundcloud import SoundCloud, User
from soundcloud.resource.like import TrackLike
from soundcloud.resource.stream import TrackStreamRepostItem


def test_valid_user_id(client: SoundCloud):
    user = client.get_user(790976431)
    assert isinstance(user, User) and user.permalink == "7x11x13"


def test_invalid_user_id(client: SoundCloud):
    user = client.get_user(0)
    assert user is None


def test_valid_username(client: SoundCloud):
    user = client.get_user_by_username("7x11x13")
    assert isinstance(user, User) and user.permalink == "7x11x13"


def test_invalid_username(client: SoundCloud):
    user = client.get_user_by_username("")
    assert user is None


def test_user_comments(client: SoundCloud):
    comment = next(client.get_user_comments(992430331))
    assert comment.body == "hi"


def test_user_conversation_messages(client: SoundCloud):
    message = next(client.get_conversation_messages(790976431, 992430331))
    assert message.content == "bye"


def test_user_conversations(client: SoundCloud):
    found = False
    for conversation in client.get_conversations(790976431):
        if conversation.last_message.content == "bye":
            found = True
    assert found


def test_user_followers(client: SoundCloud):
    found = False
    for follower in client.get_user_followers(992430331):
        if follower.permalink == "7x11x13":
            found = True
            break
    assert found


def test_user_followings(client: SoundCloud):
    following = next(client.get_user_following(992430331))
    assert following.permalink == "7x11x13"


def test_user_likes(client: SoundCloud):
    like = next(client.get_user_likes(992430331))
    assert isinstance(like, TrackLike)
    assert like.track.title == "Wan Bushi - Eurodance Vibes (part 1+2+3)"


def test_user_likes_2(client: SoundCloud):
    for like in itertools.islice(client.get_user_likes(790976431), 10):
        assert like is not None


def test_user_reposts(client: SoundCloud):
    repost = next(client.get_user_reposts(992430331))
    assert isinstance(repost, TrackStreamRepostItem)
    assert repost.track.title == "Wan Bushi - Eurodance Vibes (part 1+2+3)"


def test_user_tracks(client: SoundCloud):
    tracks = list(client.get_user_tracks(790976431))
    assert tracks[-1].title == "Wan Bushi - Eurodance Vibes (part 1+2+3)"


def test_user_toptracks(client: SoundCloud):
    tracks = list(client.get_user_popular_tracks(790976431))
    assert tracks[0].title == "Wan Bushi - Eurodance Vibes (part 1+2+3)"


def test_user_albums(client: SoundCloud):
    found = False
    for album in client.get_user_albums(211111464):
        if album.title == "Positions":
            found = True
            break
    assert found


def test_user_playlists(client: SoundCloud):
    assert next(client.get_user_playlists(992430331)).title == "playlist1"


def test_user_links(client: SoundCloud):
    user = client.get_user(992430331)
    assert user
    profiles = client.get_user_links(user.urn)
    assert profiles[0].title == "test"


def test_user_emails(client: SoundCloud):
    email = next(client.get_user_emails(790976431))
    assert email.address.endswith("gmail.com")



import datetime
from dataclasses import dataclass
from typing import Optional, Tuple, Union

from soundcloud.resource.base import BaseData
from soundcloud.resource.base_item import BaseItem
from soundcloud.resource.track import BasicTrack, MiniTrack
from soundcloud.resource.user import BasicUser, User


@dataclass
class BaseAlbumPlaylist(BaseItem):
    managed_by_feeds: bool
    set_type: str
    is_album: bool
    published_at: Optional[datetime.datetime]
    track_count: int
    tracks: Tuple[Union[BasicTrack, MiniTrack], ...]


@dataclass
class AlbumPlaylist(BaseAlbumPlaylist):
    """Playlist or album with full user info"""

    user: User


@dataclass
class BasicAlbumPlaylist(BaseAlbumPlaylist):
    """Playlist or album with partial user info"""

    user: BasicUser


@dataclass
class AlbumPlaylistNoTracks(BaseData):
    """Playlist or album with no track info"""

    artwork_url: Optional[str]
    created_at: datetime.datetime
    duration: int
    id: int
    kind: str
    last_modified: datetime.datetime
    likes_count: Optional[int]
    managed_by_feeds: bool
    permalink: str
    permalink_url: str
    public: bool
    reposts_count: Optional[int]
    secret_token: Optional[str]
    sharing: str
    title: str
    track_count: int
    uri: str
    user_id: int
    set_type: str
    is_album: bool
    published_at: Optional[datetime.datetime]
    release_date: Optional[str]
    display_date: datetime.datetime
    user: BasicUser



import datetime
from dataclasses import dataclass
from typing import Optional

from soundcloud.resource.base import BaseData
from soundcloud.resource.track import CommentTrack
from soundcloud.resource.user import BasicUser


@dataclass
class CommentSelf(BaseData):
    urn: str


@dataclass
class BasicComment(BaseData):
    """Comment without a specified track"""

    kind: str
    id: int
    body: str
    created_at: datetime.datetime
    timestamp: Optional[int]
    track_id: int
    user_id: int
    self: CommentSelf
    user: BasicUser


@dataclass
class Comment(BasicComment):
    """Comment with a specified track"""

    track: CommentTrack



from typing import Union

from soundcloud.resource.like import PlaylistLike, TrackLike
from soundcloud.resource.playlist import AlbumPlaylist
from soundcloud.resource.stream import (
    PlaylistStreamItem,
    PlaylistStreamRepostItem,
    TrackStreamItem,
    TrackStreamRepostItem,
)
from soundcloud.resource.track import Track
from soundcloud.resource.user import User

Like = Union[TrackLike, PlaylistLike]
"""Generic like"""

RepostItem = Union[TrackStreamRepostItem, PlaylistStreamRepostItem]
"""Generic repost"""

SearchItem = Union[User, Track, AlbumPlaylist]
"""Generic search result"""

StreamItem = Union[
    TrackStreamItem, PlaylistStreamItem, TrackStreamRepostItem, PlaylistStreamRepostItem
]
"""Generic feed item"""



from dataclasses import dataclass

from soundcloud.resource.base import BaseData
from soundcloud.resource.track import BasicTrack


@dataclass
class HistoryItem(BaseData):
    """Item in user's listen history"""

    played_at: int
    track: BasicTrack
    track_id: int



# ruff: noqa
"""
# soundcloud.py

[![Tests](https://github.com/7x11x13/soundcloud.py/actions/workflows/ci.yml/badge.svg)](https://github.com/7x11x13/soundcloud.py/actions/workflows/ci.yml)
[![Coverage Status](https://coveralls.io/repos/github/7x11x13/soundcloud.py/badge.svg?branch=main)](https://coveralls.io/github/7x11x13/soundcloud.py?branch=main)
[![Supported Python Versions](https://img.shields.io/badge/python-3.7%2C%203.8%2C%203.9%2C%203.10%2C%203.11%2C%203.12-blue.svg)](https://pypi.org/project/soundcloud-v2/)
[![Version](https://img.shields.io/pypi/v/soundcloud-v2.svg)](https://pypi.org/project/soundcloud-v2/)
[![License](https://img.shields.io/pypi/l/soundcloud-v2.svg)](https://pypi.org/project/soundcloud-v2/)
[![Monthly Downloads](https://pepy.tech/badge/soundcloud-v2/month)](https://pepy.tech/project/soundcloud-v2)

Python wrapper for some of the internal v2 SoundCloud API (read/GET only methods). Does not require an API key.

### Note: This is NOT the official [SoundCloud developer API](https://developers.soundcloud.com/docs/api/guide)

SoundCloud is not accepting any more application registration requests [^1] so
I made this library so developers can use SoundCloud's internal API for their projects.


[^1]: https://docs.google.com/forms/d/e/1FAIpQLSfNxc82RJuzC0DnISat7n4H-G7IsPQIdaMpe202iiHZEoso9w/closedform

## Installation

```bash
pip install soundcloud-v2
```

## Documentation

https://7x11x13.xyz/soundcloud.py

## Example

```python
from soundcloud import SoundCloud

sc = SoundCloud(auth_token="auth_token")
assert sc.is_client_id_valid()
assert sc.is_auth_token_valid()
me = sc.get_user_by_username("7x11x13")
assert me.permalink == "7x11x13"
```

## Notes on `auth_token`
Some methods require authentication in the form of an OAuth2 access token.
You can find your token in your browser cookies for SoundCloud under the name "oauth_token".
A new token will be generated each time you log out and log back in.

## Notes on `**kwargs`
All API methods have a `**kwargs` argument which you can use to pass extra, undocumented
arguments to the SoundCloud v2 API in case I missed some parameter which you find useful.
If this is the case, feel free to create an issue or pull request to document the missing
argument.

## License
[MIT](https://choosealicense.com/licenses/mit/)

"""

from soundcloud.exceptions import *
from soundcloud.exceptions import __all__ as ex_all
from soundcloud.resource import *
from soundcloud.resource import __all__ as res_all
from soundcloud.soundcloud import *
from soundcloud.soundcloud import __all__ as sc_all

__version__ = "1.6.2"

__all__ = sc_all + ex_all + res_all



from soundcloud import BasicAlbumPlaylist, SoundCloud, NoContentResponse


def test_get_playlist(client: SoundCloud):
    playlist = client.get_playlist(1326192094)
    assert (
        isinstance(playlist, BasicAlbumPlaylist)
        and playlist.user.username == "7x11x13-testing"
    )


def test_playlist_likers(client: SoundCloud):
    likers = client.get_playlist_likers(1326192094)
    found = False
    for liker in likers:
        if liker.username == "7x11x13":
            found = True
            break
    assert found


def test_playlist_reposters(client: SoundCloud):
    reposters = client.get_playlist_reposters(1326720835)
    found = False
    for reposter in reposters:
        if reposter.username == "7x11x13":
            found = True
            break
    assert found


def test_post_and_delete_playlist(client: SoundCloud):
    # POST
    playlist = client.post_playlist("private", "Playlist Test", [1032303631])
    assert (
        isinstance(playlist, BasicAlbumPlaylist)
        and playlist.title == "Playlist Test"
        and playlist.tracks[0].id == 1032303631
    )

    # DELETE
    response = client.delete_playlist(playlist.id)
    assert isinstance(response, NoContentResponse) and response.status_code == 204



import itertools

from soundcloud import SoundCloud, TrackStreamItem


def test_my_history(client: SoundCloud):
    stream = client.get_my_history()
    for item in itertools.islice(stream, 3):
        assert item is not None


def test_my_stream(client: SoundCloud):
    stream = client.get_my_stream()
    for item in itertools.islice(stream, 3):
        assert item is not None


def test_user_stream(client: SoundCloud):
    user = client.get_user_by_username("one-thousand-and-one")
    assert user
    stream = client.get_user_stream(user.id)
    item = next(stream)
    assert (
        isinstance(item, TrackStreamItem) and item.track.title == "testing - test track"
    )
    item = next(stream)
    assert (
        isinstance(item, TrackStreamItem)
        and item.track.title == "Wan Bushi - Eurodance Vibes (part 1+2+3)"
    )


def test_tag_stream(client: SoundCloud):
    stream = client.get_tag_tracks_recent("Electronic")
    for track in itertools.islice(stream, 3):
        assert track is not None



from dataclasses import dataclass
from typing import Optional, Tuple

from soundcloud.resource.base import BaseData
from soundcloud.resource.comment import BasicComment

InteractionTypeValue = str


@dataclass
class InteractionCount(BaseData):
    count: Optional[int]
    interactionTypeValueUrn: Optional[InteractionTypeValue]


@dataclass
class UserInteraction(BaseData):
    targetUrn: Optional[str]
    userInteraction: Optional[InteractionTypeValue]
    interactionCounts: Optional[Tuple[InteractionCount, ...]]
    interactionTypeUrn: Optional[str]


@dataclass
class CommentWithInteractions(BaseData):
    comment: BasicComment
    likes: int
    """Number of likes on comment"""

    liked_by_creator: bool
    """Whether comment was liked by the track creator"""

    liked_by_user: bool
    """Whether the current logged in user liked the comment"""



from dataclasses import dataclass
from typing import Tuple

from soundcloud.resource.base import BaseData


@dataclass
class Visual(BaseData):
    urn: str
    entry_time: int
    visual_url: str


@dataclass
class Visuals(BaseData):
    urn: str
    enabled: bool
    visuals: Tuple[Visual, ...]



#!/usr/bin/env python

from distutils.core import setup


def readme():
    with open("README.md", encoding="UTF-8") as f:
        return f.read()


setup(
    name="soundcloud-v2",
    version="1.6.2",
    description=(
        "Python wrapper for the internal v2 SoundCloud API.Does not require an API key."
    ),
    long_description=readme(),
    long_description_content_type="text/markdown",
    author="7x11x13",
    author_email="x7x11x13@gmail.com",
    url="https://github.com/7x11x13/soundcloud.py",
    packages=["soundcloud", "soundcloud.resource"],
    package_data={"soundcloud": ["py.typed"]},
    install_requires=[
        "dacite>=1.8.1",
        "python-dateutil>=2.8.2",
        "requests",
        "typing_extensions; python_version<'3.8'",
    ],
    extras_require={
        "dev": [
            "coveralls",
            "pytest",
            "pytest-dotenv",
            "types-python-dateutil",
            "types-requests",
            "mypy",
            "ruff",
        ],
        "docs": ["pdoc"],
    },
    classifiers=[
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.7",
    project_urls={"Bug Tracker": "https://github.com/7x11x13/soundcloud.py/issues"},
)



from soundcloud import AlbumPlaylist, SoundCloud, Track, User


def test_search_all(client: SoundCloud):
    user = next(client.search("Ariana Grande"))
    assert isinstance(user, User) and user.permalink == "arianagrande"


def test_search_albums(client: SoundCloud):
    album = next(client.search_albums("Ariana Grande - Positions"))
    assert isinstance(album, AlbumPlaylist) and album.user.username == "Ariana Grande"


def test_search_playlists(client: SoundCloud):
    playlist = next(client.search_playlists("we do a little music"))
    assert (
        isinstance(playlist, AlbumPlaylist)
        and playlist.title == "We Do A Little Music [incomplete]"
    )


def test_search_tracks(client: SoundCloud):
    track = next(client.search_tracks("34+35"))
    assert isinstance(track, Track) and track.user.username == "Ariana Grande"


def test_search_users(client: SoundCloud):
    user = next(client.search_users("namasenda"))
    assert isinstance(user, User) and user.permalink == "namasenda"



from dataclasses import dataclass

from soundcloud.resource.base import BaseData


@dataclass
class NoContentResponse(BaseData):
    """Response with no content, mainly for DELETE requests"""

    status_code: int



from dataclasses import dataclass
from typing import Optional

from soundcloud.resource.base import BaseData


@dataclass
class WebProfile(BaseData):
    url: str
    network: str
    title: str
    username: Optional[str]



import datetime
from dataclasses import dataclass
from typing import Optional

from soundcloud.resource.base import BaseData
from soundcloud.resource.playlist import BasicAlbumPlaylist
from soundcloud.resource.track import BasicTrack
from soundcloud.resource.user import BasicUser


@dataclass
class BaseStreamItem(BaseData):
    created_at: datetime.datetime
    type: str
    user: BasicUser
    uuid: str
    caption: Optional[str]


@dataclass
class Reposted(BaseData):
    target_urn: str
    user_urn: str
    caption: Optional[str]


@dataclass
class BaseStreamRepostItem(BaseStreamItem):
    reposted: Optional[Reposted]


@dataclass
class TrackStreamItem(BaseStreamItem):
    """Track post in user's feed"""

    track: BasicTrack


@dataclass
class TrackStreamRepostItem(BaseStreamRepostItem):
    """Track repost in user's feed"""

    track: BasicTrack


@dataclass
class PlaylistStreamItem(BaseStreamItem):
    """Album or playlist post in user's feed"""

    playlist: BasicAlbumPlaylist


@dataclass
class PlaylistStreamRepostItem(BaseStreamRepostItem):
    """Album or playlist repost in user's feed"""

    playlist: BasicAlbumPlaylist



import itertools
import sys
import re
from typing import Dict, Generator, List, Optional

if sys.version_info < (3, 8):
    from typing_extensions import Literal
else:
    from typing import Literal

import requests
from requests import HTTPError

from soundcloud.exceptions import ClientIDGenerationError
from soundcloud.requests import (
    MeHistoryRequest,
    MeRequest,
    MeStreamRequest,
    PlaylistLikersRequest,
    PlaylistRepostersRequest,
    PlaylistRequest,
    PostPlaylistRequest,
    DeletePlaylistRequest,
    ResolveRequest,
    SearchAlbumsRequest,
    SearchPlaylistsRequest,
    SearchRequest,
    SearchTracksRequest,
    SearchUsersRequest,
    TagRecentTracksRequest,
    TrackAlbumsRequest,
    TrackCommentsRequest,
    TrackLikersRequest,
    TrackOriginalDownloadRequest,
    TrackPlaylistsRequest,
    TrackRelatedRequest,
    TrackRepostersRequest,
    TrackRequest,
    TracksRequest,
    UserAlbumsRequest,
    UserCommentsRequest,
    UserConversationMessagesRequest,
    UserConversationsRequest,
    UserConversationsUnreadRequest,
    UserEmailsRequest,
    UserFeaturedProfilesRequest,
    UserFollowersRequest,
    UserFollowingsRequest,
    UserInteractionsQueryParams,
    UserInteractionsRequest,
    UserLikesRequest,
    UserPlaylistsRequest,
    UserRelatedArtistsRequest,
    UserRepostsRequest,
    UserRequest,
    UserStreamRequest,
    UserToptracksRequest,
    UserTracksRequest,
    UserWebProfilesRequest,
)
from soundcloud.resource.graphql import CommentWithInteractions
from soundcloud.resource.history import HistoryItem

from .resource.aliases import Like, RepostItem, SearchItem, StreamItem
from .resource.comment import BasicComment, Comment
from .resource.conversation import Conversation
from .resource.message import Message
from .resource.playlist import AlbumPlaylist, BasicAlbumPlaylist
from .resource.track import BasicTrack, Track
from .resource.user import User, UserEmail
from .resource.web_profile import WebProfile
from .resource.response import NoContentResponse


class SoundCloud:
    """
    SoundCloud v2 API client
    """

    _DEFAULT_USER_AGENT = (
        "Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:93.0) Gecko/20100101 Firefox/93.0"
    )
    _ASSETS_SCRIPTS_REGEX = re.compile(
        r"src=\"(https:\/\/a-v2\.sndcdn\.com/assets/.*\.js)\""
    )
    _CLIENT_ID_REGEX = re.compile(r"client_id:\"([^\"]+)\"")
    client_id: str
    """SoundCloud client ID. Needed for all requests."""
    _user_agent: str
    _auth_token: Optional[str]
    _authorization: Optional[str]

    def __init__(
        self,
        client_id: Optional[str] = None,
        auth_token: Optional[str] = None,
        user_agent: str = _DEFAULT_USER_AGENT,
    ) -> None:
        if not client_id:
            client_id = self.generate_client_id()

        self.client_id = client_id
        self._user_agent = user_agent
        self._auth_token = None
        self._authorization = None
        self.auth_token = auth_token

    @property
    def auth_token(self) -> Optional[str]:
        """SoundCloud auth token. Only needed for some requests."""
        return self._auth_token

    @auth_token.setter
    def auth_token(self, new_auth_token: Optional[str]) -> None:
        if new_auth_token is not None:
            if new_auth_token.startswith("OAuth"):
                new_auth_token = new_auth_token.split()[-1]
        self._authorization = f"OAuth {new_auth_token}" if new_auth_token else None
        self._auth_token = new_auth_token

    @auth_token.deleter
    def auth_token(self):
        self.auth_token = None

    def _get_default_headers(self) -> Dict[str, str]:
        return {"User-Agent": self._user_agent}

    @classmethod
    def generate_client_id(cls) -> str:
        """Generates a SoundCloud client ID

        Raises:
            ClientIDGenerationError: Client ID could not be generated.

        Returns:
            str: Valid client ID
        """
        r = requests.get("https://soundcloud.com")
        r.raise_for_status()
        matches = cls._ASSETS_SCRIPTS_REGEX.findall(r.text)
        if not matches:
            raise ClientIDGenerationError("No asset scripts found")
        for url in matches:
            r = requests.get(url)
            r.raise_for_status()
            client_id = cls._CLIENT_ID_REGEX.search(r.text)
            if client_id:
                return client_id.group(1)
        raise ClientIDGenerationError(f"Could not find client_id in script '{url}'")

    def is_client_id_valid(self) -> bool:
        """
        Checks if current client_id is valid
        """
        try:
            TrackRequest(self, track_id=1032303631, use_auth=False)
            return True
        except HTTPError as err:
            if err.response.status_code == 401:
                return False
            else:
                raise

    def is_auth_token_valid(self) -> bool:
        """
        Checks if current auth_token is valid
        """
        try:
            MeRequest(self)
            return True
        except HTTPError as err:
            if err.response.status_code == 401:
                return False
            else:
                raise

    def get_me(self) -> Optional[User]:
        """
        Gets the user associated with client's auth token
        """
        return MeRequest(self)

    def get_my_history(self, **kwargs) -> Generator[HistoryItem, None, None]:
        """
        Returns the stream of recently listened tracks
        for the client's auth token
        """
        return MeHistoryRequest(self, **kwargs)

    def get_my_stream(self, **kwargs) -> Generator[StreamItem, None, None]:
        """
        Returns the stream of recent uploads/reposts
        for the client's auth token
        """
        return MeStreamRequest(self, **kwargs)

    def resolve(self, url: str) -> Optional[SearchItem]:
        """
        Returns the resource at the given URL if it
        exists, otherwise return None
        """
        return ResolveRequest(self, url=url)

    def search(self, query: str, **kwargs) -> Generator[SearchItem, None, None]:
        """
        Search for users, tracks, and playlists
        """
        return SearchRequest(self, q=query, **kwargs)

    def search_albums(
        self, query: str, **kwargs
    ) -> Generator[AlbumPlaylist, None, None]:
        """
        Search for albums (not playlists)
        """
        return SearchAlbumsRequest(self, q=query, **kwargs)

    def search_playlists(
        self, query: str, **kwargs
    ) -> Generator[AlbumPlaylist, None, None]:
        """
        Search for playlists
        """
        return SearchPlaylistsRequest(self, q=query, **kwargs)

    def search_tracks(self, query: str, **kwargs) -> Generator[Track, None, None]:
        """
        Search for tracks
        """
        return SearchTracksRequest(self, q=query, **kwargs)

    def search_users(self, query: str, **kwargs) -> Generator[User, None, None]:
        """
        Search for users
        """
        return SearchUsersRequest(self, q=query, **kwargs)

    def get_tag_tracks_recent(self, tag: str, **kwargs) -> Generator[Track, None, None]:
        """
        Get most recent tracks for this tag
        """
        return TagRecentTracksRequest(self, tag=tag, **kwargs)

    def get_playlist(self, playlist_id: int) -> Optional[BasicAlbumPlaylist]:
        """
        Returns the playlist with the given playlist_id.
        If the ID is invalid, return None
        """
        return PlaylistRequest(self, playlist_id=playlist_id)

    def post_playlist(
        self, sharing: Literal["private", "public"], title: str, tracks: List[int]
    ) -> Optional[BasicAlbumPlaylist]:
        """
        Create a new playlist
        """
        body = {"playlist": {"sharing": sharing, "title": title, "tracks": tracks}}
        return PostPlaylistRequest(self, body=body)

    def delete_playlist(self, playlist_id: int) -> Optional[NoContentResponse]:
        """
        Delete a playlist
        """
        return DeletePlaylistRequest(self, playlist_id=playlist_id)

    def get_playlist_likers(
        self, playlist_id: int, **kwargs
    ) -> Generator[User, None, None]:
        """
        Get people who liked this playlist
        """
        return PlaylistLikersRequest(self, playlist_id=playlist_id, **kwargs)

    def get_playlist_reposters(
        self, playlist_id: int, **kwargs
    ) -> Generator[User, None, None]:
        """
        Get people who reposted this playlist
        """
        return PlaylistRepostersRequest(self, playlist_id=playlist_id, **kwargs)

    def get_track(self, track_id: int) -> Optional[BasicTrack]:
        """
        Returns the track with the given track_id.
        If the ID is invalid, return None
        """
        return TrackRequest(self, track_id=track_id)

    def get_tracks(
        self,
        track_ids: List[int],
        playlistId: Optional[int] = None,
        playlistSecretToken: Optional[str] = None,
        **kwargs,
    ) -> List[BasicTrack]:
        """
        Returns the tracks with the given track_ids.
        Can be used to get track info for hidden tracks in a hidden playlist.
        """
        if playlistId is not None:
            kwargs["playlistId"] = playlistId
        if playlistSecretToken is not None:
            kwargs["playlistSecretToken"] = playlistSecretToken
        return TracksRequest(
            self, ids=",".join([str(id) for id in track_ids]), **kwargs
        )

    def get_track_albums(
        self, track_id: int, **kwargs
    ) -> Generator[BasicAlbumPlaylist, None, None]:
        """
        Get albums that this track is in
        """
        return TrackAlbumsRequest(self, track_id=track_id, **kwargs)

    def get_track_playlists(
        self, track_id: int, **kwargs
    ) -> Generator[BasicAlbumPlaylist, None, None]:
        """
        Get playlists that this track is in
        """
        return TrackPlaylistsRequest(self, track_id=track_id, **kwargs)

    def get_track_comments(
        self, track_id: int, threaded: int = 0, **kwargs
    ) -> Generator[BasicComment, None, None]:
        """
        Get comments on this track
        """
        return TrackCommentsRequest(
            self, track_id=track_id, threaded=threaded, **kwargs
        )

    def get_track_comments_with_interactions(
        self, track_id: int, threaded: int = 0, **kwargs
    ) -> Generator[CommentWithInteractions, None, None]:
        """
        Get comments on this track with interaction data. Requires authentication.
        """
        track = self.get_track(track_id)
        if not track:
            return
        track_urn = track.urn
        creator_urn = track.user.urn
        comments = self.get_track_comments(track_id, threaded, **kwargs)
        while True:
            chunk = list(itertools.islice(comments, 10))
            if not chunk:
                return
            comment_urns = [comment.self.urn for comment in chunk]
            result = UserInteractionsRequest(
                self,
                UserInteractionsQueryParams(
                    creator_urn,
                    "sc:interactiontype:reaction",
                    track_urn,
                    comment_urns,
                ),
            )
            if not result:
                return
            comments_with_interactions = []
            for comment, user_interactions, creator_interactions in zip(
                chunk, result.user, result.creator
            ):
                assert user_interactions.interactionCounts is not None
                likes = list(
                    filter(
                        lambda x: x.interactionTypeValueUrn
                        == "sc:interactiontypevalue:like",
                        user_interactions.interactionCounts,
                    )
                )
                num_likes = likes[0].count if likes else 0
                comments_with_interactions.append(
                    CommentWithInteractions(
                        comment=comment,
                        likes=num_likes or 0,
                        liked_by_creator=creator_interactions.userInteraction
                        == "sc:interactiontypevalue:like",
                        liked_by_user=user_interactions.userInteraction
                        == "sc:interactiontypevalue:like",
                    )
                )
            yield from comments_with_interactions

    def get_track_likers(self, track_id: int, **kwargs) -> Generator[User, None, None]:
        """
        Get users who liked this track
        """
        return TrackLikersRequest(self, track_id=track_id, **kwargs)

    def get_track_related(
        self, track_id: int, **kwargs
    ) -> Generator[BasicTrack, None, None]:
        """
        Get related tracks
        """
        return TrackRelatedRequest(self, track_id=track_id, **kwargs)

    def get_track_reposters(
        self, track_id: int, **kwargs
    ) -> Generator[User, None, None]:
        """
        Get users who reposted this track
        """
        return TrackRepostersRequest(self, track_id=track_id, **kwargs)

    def get_track_original_download(
        self, track_id: int, token: Optional[str] = None
    ) -> Optional[str]:
        """
        Get track original download link. If track is private,
        requires secret token to be provided (last part of secret URL).
        Requires authentication.
        """
        if token is not None:
            download = TrackOriginalDownloadRequest(
                self, track_id=track_id, secret_token=token
            )
        else:
            download = TrackOriginalDownloadRequest(self, track_id=track_id)
        if download is None:
            return None
        else:
            return download.redirectUri

    def get_user(self, user_id: int) -> Optional[User]:
        """
        Returns the user with the given user_id.
        If the ID is invalid, return None
        """
        return UserRequest(self, user_id=user_id)

    def get_user_by_username(self, username: str) -> Optional[User]:
        """
        Returns the user with the given username.
        If the username is invalid, return None
        """
        resource = self.resolve(f"https://soundcloud.com/{username}")
        if resource and isinstance(resource, User):
            return resource
        else:
            return None

    def get_user_comments(
        self, user_id: int, **kwargs
    ) -> Generator[Comment, None, None]:
        """
        Get comments by this user
        """
        return UserCommentsRequest(self, user_id=user_id, **kwargs)

    def get_conversation_messages(
        self, user_id: int, conversation_id: int, **kwargs
    ) -> Generator[Message, None, None]:
        """
        Get messages in this conversation
        """
        return UserConversationMessagesRequest(
            self, user_id=user_id, conversation_id=conversation_id, **kwargs
        )

    def get_conversations(
        self, user_id: int, **kwargs
    ) -> Generator[Conversation, None, None]:
        """
        Get conversations including this user
        """
        return UserConversationsRequest(self, user_id=user_id, **kwargs)

    def get_unread_conversations(
        self, user_id: int, **kwargs
    ) -> Generator[Conversation, None, None]:
        """
        Get conversations unread by this user
        """
        return UserConversationsUnreadRequest(self, user_id=user_id, **kwargs)

    def get_user_emails(
        self, user_id: int, **kwargs
    ) -> Generator[UserEmail, None, None]:
        """
        Get user's email addresses. Requires authentication.
        """
        return UserEmailsRequest(self, user_id=user_id, **kwargs)

    def get_user_featured_profiles(
        self, user_id: int, **kwargs
    ) -> Generator[User, None, None]:
        """
        Get profiles featured by this user
        """
        return UserFeaturedProfilesRequest(self, user_id=user_id, **kwargs)

    def get_user_followers(self, user_id: int, **kwargs) -> Generator[User, None, None]:
        """
        Get user's followers
        """
        return UserFollowersRequest(self, user_id=user_id, **kwargs)

    def get_user_following(self, user_id: int, **kwargs) -> Generator[User, None, None]:
        """
        Get users this user is following
        """
        return UserFollowingsRequest(self, user_id=user_id, **kwargs)

    def get_user_likes(self, user_id: int, **kwargs) -> Generator[Like, None, None]:
        """
        Get likes by this user
        """
        return UserLikesRequest(self, user_id=user_id, **kwargs)

    def get_user_related_artists(
        self, user_id: int, **kwargs
    ) -> Generator[User, None, None]:
        """
        Get artists related to this user
        """
        return UserRelatedArtistsRequest(self, user_id=user_id, **kwargs)

    def get_user_reposts(
        self, user_id: int, **kwargs
    ) -> Generator[RepostItem, None, None]:
        """
        Get reposts by this user
        """
        return UserRepostsRequest(self, user_id=user_id, **kwargs)

    def get_user_stream(
        self, user_id: int, **kwargs
    ) -> Generator[StreamItem, None, None]:
        """
        Returns generator of track uploaded by given user and
        reposts by this user
        """
        return UserStreamRequest(self, user_id=user_id, **kwargs)

    def get_user_tracks(
        self, user_id: int, **kwargs
    ) -> Generator[BasicTrack, None, None]:
        """
        Get tracks uploaded by this user
        """
        return UserTracksRequest(self, user_id=user_id, **kwargs)

    def get_user_popular_tracks(
        self, user_id: int, **kwargs
    ) -> Generator[BasicTrack, None, None]:
        """
        Get popular tracks uploaded by this user
        """
        return UserToptracksRequest(self, user_id=user_id, **kwargs)

    def get_user_albums(
        self, user_id: int, **kwargs
    ) -> Generator[BasicAlbumPlaylist, None, None]:
        """
        Get albums uploaded by this user
        """
        return UserAlbumsRequest(self, user_id=user_id, **kwargs)

    def get_user_playlists(
        self, user_id: int, **kwargs
    ) -> Generator[BasicAlbumPlaylist, None, None]:
        """
        Get playlists uploaded by this user
        """
        return UserPlaylistsRequest(self, user_id=user_id, **kwargs)

    def get_user_links(self, user_urn: str, **kwargs) -> List[WebProfile]:
        """
        Get links in this user's description
        """
        return UserWebProfilesRequest(self, user_urn=user_urn, **kwargs)


__all__ = ["SoundCloud"]



import datetime
from dataclasses import dataclass
from typing import Optional, Tuple

from soundcloud.resource.base import BaseData
from soundcloud.resource.visuals import Visuals


@dataclass
class Product(BaseData):
    id: str


@dataclass
class CreatorSubscription(BaseData):
    product: Product


@dataclass
class Badges(BaseData):
    """User badges"""

    pro: bool
    pro_unlimited: bool
    verified: bool


@dataclass
class BasicUser(BaseData):
    """User with partial information"""

    avatar_url: str
    first_name: str
    followers_count: int
    full_name: str
    id: int
    kind: str
    last_modified: datetime.datetime
    last_name: str
    permalink: str
    permalink_url: str
    uri: str
    urn: str
    username: str
    verified: bool
    city: Optional[str]
    country_code: Optional[str]
    badges: Badges
    station_urn: Optional[str]
    station_permalink: Optional[str]


@dataclass
class User(BasicUser):
    """User with full information"""

    comments_count: int
    created_at: Optional[datetime.datetime]
    creator_subscriptions: Tuple[CreatorSubscription, ...]
    creator_subscription: CreatorSubscription
    description: Optional[str]
    followings_count: int
    groups_count: int
    likes_count: Optional[int]
    playlist_likes_count: Optional[int]
    playlist_count: int
    reposts_count: Optional[int]
    track_count: int
    visuals: Optional[Visuals]


@dataclass
class MissingUser(BaseData):
    """Deleted user"""

    id: int
    kind: str


@dataclass
class UserStatus(BaseData):
    status: str
    timestamp: str


@dataclass
class UserEmail(BaseData):
    """Email address associated with a user"""

    address: str
    confirmed: bool
    id: int
    kind: str
    last_modified: datetime.datetime
    primary: bool
    urn: str
    user_id: str



import string
from dataclasses import asdict, dataclass
import sys
from typing import (
    TYPE_CHECKING,
    Any,
    ClassVar,
    Dict,
    Generator,
    Generic,
    List,
    Optional,
    Tuple,
    Type,
    TypeVar,
    Union,
)

import requests

from soundcloud.resource.aliases import Like, RepostItem, SearchItem, StreamItem
from soundcloud.resource.base import BaseData
from soundcloud.resource.comment import BasicComment, Comment
from soundcloud.resource.conversation import Conversation
from soundcloud.resource.download import OriginalDownload
from soundcloud.resource.graphql import UserInteraction
from soundcloud.resource.history import HistoryItem
from soundcloud.resource.message import Message
from soundcloud.resource.playlist import AlbumPlaylist, BasicAlbumPlaylist
from soundcloud.resource.response import NoContentResponse
from soundcloud.resource.track import BasicTrack, Track
from soundcloud.resource.user import User, UserEmail
from soundcloud.resource.web_profile import WebProfile

if TYPE_CHECKING:
    from soundcloud.soundcloud import SoundCloud

if sys.version_info >= (3, 8):
    from typing import Protocol
else:
    from typing_extensions import Protocol

try:
    from typing import get_args, get_origin  # type: ignore[attr-defined]
except ImportError:
    # get_args and get_origin for version < 3.8
    def get_args(tp):  # type: ignore[misc]
        return getattr(tp, "__args__", ())

    def get_origin(tp):  # type: ignore[no-redef]
        return getattr(tp, "__origin__", None)


from urllib.parse import parse_qs, urljoin, urlparse


def _convert_dict(d, return_type: Type[BaseData]):
    union = get_origin(return_type) is Union
    if union:
        for t in get_args(return_type):
            try:
                return t.from_dict(d)
            except Exception:
                pass
    else:
        return return_type.from_dict(d)
    raise ValueError(f"Could not convert {d} to type {return_type}")


T = TypeVar("T", bound=BaseData)


@dataclass
class Request(Generic[T]):
    base = "https://api-v2.soundcloud.com"
    format_url: str
    return_type: Type[T]
    method: str = "GET"

    def _format_url_and_remove_params(self, kwargs: dict) -> str:
        format_args = {
            tup[1]
            for tup in string.Formatter().parse(self.format_url)
            if tup[1] is not None
        }
        args = {}
        for k in list(kwargs.keys()):
            if k in format_args:
                args[k] = kwargs.pop(k)
        return self.base + self.format_url.format(**args)

    def __call__(
        self,
        client: "SoundCloud",
        use_auth: bool = True,
        body: Optional[dict] = None,
        **kwargs,
    ) -> Optional[T]:
        """
        Requests the resource at the given url with
        parameters given by kwargs. Converts the resource
        to type T and returns it. If the
        resource does not exist, returns None
        """
        resource_url = self._format_url_and_remove_params(kwargs)
        params = kwargs
        params["client_id"] = client.client_id
        headers = client._get_default_headers()

        if use_auth and client._authorization is not None:
            headers["Authorization"] = client._authorization

        with requests.request(
            self.method, resource_url, json=body, headers=headers, params=params
        ) as r:
            if r.status_code in (400, 404, 500):
                return None
            r.raise_for_status()

        if self.return_type == NoContentResponse:
            return NoContentResponse(r.status_code)  # type: ignore[return-value]
        return _convert_dict(r.json(), self.return_type)


@dataclass
class CollectionRequest(Request, Generic[T]):
    """
    Yields resources from the given url with
    parameters given by kwargs. Converts the resources
    to type T before yielding
    """

    def __call__(
        self,
        client: "SoundCloud",
        use_auth: bool = True,
        body: Optional[dict] = None,
        offset: Optional[str] = None,
        limit: Optional[int] = None,
        **kwargs,
    ) -> Generator[T, None, None]:
        resource_url = self._format_url_and_remove_params(kwargs)
        params = kwargs
        params["client_id"] = client.client_id
        if offset is not None:
            params["offset"] = offset
        if limit is not None:
            params["limit"] = limit
        headers = client._get_default_headers()
        if use_auth and client._authorization is not None:
            headers["Authorization"] = client._authorization
        while resource_url:
            with requests.get(resource_url, params=params, headers=headers) as r:
                if r.status_code in (400, 404, 500):
                    return
                r.raise_for_status()
                data = r.json()
                for resource in data["collection"]:
                    yield _convert_dict(resource, self.return_type)
                resource_url = data.get("next_href", None)
                parsed = urlparse(resource_url)
                params = parse_qs(parsed.query)
                params["client_id"] = [
                    client.client_id
                ]  # next_href doesn't contain client_id
                resource_url = urljoin(resource_url, parsed.path)


@dataclass
class ListRequest(Request, Generic[T]):
    """
    Requests the resource list at the given url with
    parameters given by kwargs. Converts the resources
    to type T and returns them.
    """

    def __call__(
        self, client: "SoundCloud", use_auth=True, body: Optional[dict] = None, **kwargs
    ) -> List[T]:
        resource_url = self._format_url_and_remove_params(kwargs)
        params = kwargs
        params["client_id"] = client.client_id
        headers = client._get_default_headers()
        if use_auth and client._authorization is not None:
            headers["Authorization"] = client._authorization
        resources = []
        with requests.get(resource_url, params=params, headers=headers) as r:
            if r.status_code in (400, 404, 500):
                return []
            r.raise_for_status()
            for resource in r.json():
                resources.append(_convert_dict(resource, self.return_type))
        return resources


class DataclassInstance(Protocol):
    __dataclass_fields__: ClassVar[Dict[str, Any]]


Q = TypeVar("Q", bound=DataclassInstance)


@dataclass
class GraphQLRequest(Generic[Q, T]):
    base = "https://graph.soundcloud.com/graphql"
    operation_name: str
    query_arg_type: Type[Q]
    return_type: Type[T]
    query_template_str: str

    def __call__(
        self,
        client: "SoundCloud",
        query_args: Q,
        use_auth=True,
    ) -> Optional[T]:
        params = {}
        params["client_id"] = client.client_id
        headers = client._get_default_headers()
        headers["Apollographql-Client-Name"] = "v2"
        if use_auth and client._authorization is not None:
            headers["Authorization"] = client._authorization

        data = {
            "operationName": self.operation_name,
            "query": self.query_template_str,
            "variables": asdict(query_args),
        }

        with requests.post(self.base, json=data, params=params, headers=headers) as r:
            if r.status_code in (400, 404, 500):
                return None
            r.raise_for_status()
            return _convert_dict(r.json()["data"], self.return_type)


"""
v2 endpoints
"""


MeRequest = Request[User]("/me", User)
MeHistoryRequest = CollectionRequest[HistoryItem](
    "/me/play-history/tracks", HistoryItem
)
MeStreamRequest = CollectionRequest[StreamItem](
    "/stream",
    StreamItem,  # type: ignore[arg-type]
)
ResolveRequest = Request[SearchItem]("/resolve", SearchItem)  # type: ignore[arg-type]
SearchRequest = CollectionRequest[SearchItem](
    "/search",
    SearchItem,  # type: ignore[arg-type]
)
SearchAlbumsRequest = CollectionRequest[AlbumPlaylist](
    "/search/albums", AlbumPlaylist
)  # ?filter.genre_or_tag
SearchPlaylistsRequest = CollectionRequest[AlbumPlaylist](
    "/search/playlists_without_albums", AlbumPlaylist
)
SearchTracksRequest = CollectionRequest[Track](
    "/search/tracks", Track
)  # ?filter.created_at&filter.duration&filter.license
SearchUsersRequest = CollectionRequest[User]("/search/users", User)  # ?filter.place
TagRecentTracksRequest = CollectionRequest[Track]("/recent-tracks/{tag}", Track)
PlaylistRequest = Request[BasicAlbumPlaylist](
    "/playlists/{playlist_id}", BasicAlbumPlaylist
)
PostPlaylistRequest = Request[BasicAlbumPlaylist](
    "/playlists", BasicAlbumPlaylist, method="POST"
)
DeletePlaylistRequest = Request[NoContentResponse](
    "/playlists/{playlist_id}", NoContentResponse, method="DELETE"
)
PlaylistLikersRequest = CollectionRequest[User]("/playlists/{playlist_id}/likers", User)
PlaylistRepostersRequest = CollectionRequest[User](
    "/playlists/{playlist_id}/reposters", User
)
TrackRequest = Request[BasicTrack]("/tracks/{track_id}", BasicTrack)
TracksRequest = ListRequest[BasicTrack]("/tracks", BasicTrack)
TrackAlbumsRequest = CollectionRequest[BasicAlbumPlaylist](
    "/tracks/{track_id}/albums", BasicAlbumPlaylist
)  # (can be representation=mini)
TrackPlaylistsRequest = CollectionRequest[BasicAlbumPlaylist](
    "/tracks/{track_id}/playlists_without_albums", BasicAlbumPlaylist
)  # (can be representation=mini)
TrackCommentsRequest = CollectionRequest[BasicComment](
    "/tracks/{track_id}/comments", BasicComment
)
TrackLikersRequest = CollectionRequest[User]("/tracks/{track_id}/likers", User)
TrackRelatedRequest = CollectionRequest[BasicTrack](
    "/tracks/{track_id}/related", BasicTrack
)
TrackRepostersRequest = CollectionRequest[User]("/tracks/{track_id}/reposters", User)
TrackOriginalDownloadRequest = Request[OriginalDownload](
    "/tracks/{track_id}/download", OriginalDownload
)
UserRequest = Request[User]("/users/{user_id}", User)
UserCommentsRequest = CollectionRequest[Comment]("/users/{user_id}/comments", Comment)
UserConversationMessagesRequest = CollectionRequest[Message](
    "/users/{user_id}/conversations/{conversation_id}/messages", Message
)
UserConversationsRequest = CollectionRequest[Conversation](
    "/users/{user_id}/conversations", Conversation
)
UserConversationsUnreadRequest = CollectionRequest[Conversation](
    "/users/{user_id}/conversations/unread", Conversation
)
UserEmailsRequest = CollectionRequest[UserEmail]("/users/{user_id}/emails", UserEmail)
UserFeaturedProfilesRequest = CollectionRequest[User](
    "/users/{user_id}/featured-profiles", User
)
UserFollowersRequest = CollectionRequest[User]("/users/{user_id}/followers", User)
UserFollowingsRequest = CollectionRequest[User]("/users/{user_id}/followings", User)
UserLikesRequest = CollectionRequest[Like](
    "/users/{user_id}/likes",
    Like,  # type: ignore[arg-type]
)
UserRelatedArtistsRequest = CollectionRequest[User](
    "/users/{user_id}/relatedartists", User
)
UserRepostsRequest = CollectionRequest[RepostItem](
    "/stream/users/{user_id}/reposts",
    RepostItem,  # type: ignore[arg-type]
)
UserStreamRequest = CollectionRequest[StreamItem](
    "/stream/users/{user_id}",
    StreamItem,  # type: ignore[arg-type]
)
UserTracksRequest = CollectionRequest[BasicTrack]("/users/{user_id}/tracks", BasicTrack)
UserToptracksRequest = CollectionRequest[BasicTrack](
    "/users/{user_id}/toptracks", BasicTrack
)
UserAlbumsRequest = CollectionRequest[BasicAlbumPlaylist](
    "/users/{user_id}/albums", BasicAlbumPlaylist
)  # (can be representation=mini)
UserPlaylistsRequest = CollectionRequest[BasicAlbumPlaylist](
    "/users/{user_id}/playlists_without_albums", BasicAlbumPlaylist
)  # (can be representation=mini)
UserWebProfilesRequest = ListRequest[WebProfile](
    "/users/{user_urn}/web-profiles", WebProfile
)


"""
graphql endpoints
"""


@dataclass
class UserInteractionsQueryResult(BaseData):
    user: Tuple[UserInteraction, ...]
    creator: Tuple[UserInteraction, ...]


@dataclass
class UserInteractionsQueryParams:
    createdByProfileUrn: str
    interactionTypeUrn: str
    parentUrn: str
    targetUrns: List[str]


UserInteractionsRequest = GraphQLRequest[
    UserInteractionsQueryParams, UserInteractionsQueryResult
](
    "UserInteractions",
    UserInteractionsQueryParams,
    UserInteractionsQueryResult,
    (
        "query UserInteractions(\n"
        "   $parentUrn: String!\n"
        "   $interactionTypeUrn: String!\n"
        "   $targetUrns: [String!]!\n"
        "   $createdByProfileUrn: String\n"
        ") {\n"
        "   user: userInteractions(\n"
        "       parentUrn: $parentUrn\n"
        "       interactionTypeUrn: $interactionTypeUrn\n"
        "       targetUrns: $targetUrns\n"
        "   ) {\n"
        "       interactionCounts {\n"
        "           count\n"
        "           interactionTypeValueUrn\n"
        "       }\n"
        "       interactionTypeUrn\n"
        "       targetUrn\n"
        "       userInteraction\n"
        "   }\n"
        "\n"
        "   creator: userInteractions(\n"
        "       parentUrn: $parentUrn\n"
        "       interactionTypeUrn: $interactionTypeUrn\n"
        "       targetUrns: $targetUrns\n"
        "       createdByProfileUrn: $createdByProfileUrn\n"
        "   ) {\n"
        "       interactionCounts {\n"
        "           count\n"
        "           interactionTypeValueUrn\n"
        "       }\n"
        "       interactionTypeUrn\n"
        "       targetUrn\n"
        "       userInteraction\n"
        "   }\n"
        "}"
    ),
)
